import pandas as pd
import numpy as np

def data_transformation(df, df_municipalities_counties, df_macro):
    df.rename(columns={'Aar':'Year',
                   'Maaned':'Month',
                   'KundeID':'Customer_ID',
                   'Inntekt':'Income',
                   'StorrelseHusholdning':'Size_Household',
                   'Bosted':'Municipality',
                   'NedbetalingsPlan':'Repayment_Plan',
                   'SumNedbetalingsLaan':'Repayment_Loan_Balance',
                   'VektetSnittRente':'Weighted_Average_Interest_Rate',
                   'SumSaldoBoligkreditt':'Boligkreditt_Balance',
                   'SumRammeBoligkreditt':'Boligkreditt_Credit_Limit',
                   'SumSeniorLaan':'Seniorlaan_Balance',
                   'AnnenInternGjeld':'Other_Internal_Debt',
                   'AnnenEkstGjeld':'Other_External_Debt',
                   'Alder':'Age',
                   'Dato':'Date',
                   'Kjonn_Dummy':'Sex_Dummy',
                   'Kredittkort_Dummy':'Credit_Card_Dummy',
                   'Innskudd_Dummy':'Deposit_Dummy',
                   'last_row_indicator':'Churned',
                   'MonthDifference':'Customer_Relationship_Duration',
                   'last_row_indicator_3lag':'Churned_3_Lag',
                   'last_row_indicator_6lag':'Churned_6_Lag'},inplace=True)

    table = {'Original dataset':len(df)}

    # Merge with df_municipalities_counties
    df_municipalities_counties['Municipality'] = df_municipalities_counties['Municipality'].str.upper()
    df['Municipality'] = df['Municipality'].str.upper()
    df = pd.merge(df, df_municipalities_counties, on='Municipality', how='left')

    # Merge with data_makro
    df['Month'] = df['Month'].str.strip()
    df = pd.merge(df, df_macro, on=['Year', 'Month'], how='left')

    # Replacing nan/NULL-values
    # Replace NULL with 0
    df[['Repayment_Loan_Balance','Repayment_Plan','Boligkreditt_Balance','Boligkreditt_Credit_Limit','Seniorlaan_Balance','Other_Internal_Debt','Other_External_Debt']] = df[['Repayment_Loan_Balance','Repayment_Plan','Boligkreditt_Balance','Boligkreditt_Credit_Limit','Seniorlaan_Balance','Other_Internal_Debt','Other_External_Debt']].fillna(0)

    # Replace NULL-values in "County" with "Other_County"
    df['County'] = df['County'].fillna('Other_County')

    # Fill the null values in 'Weighted_Average_Interest_Rate' if 'Seniorlaan_Balance' is larger than 0
    df.loc[df['Seniorlaan_Balance'] > 0, 'Weighted_Average_Interest_Rate'] = df.loc[df['Seniorlaan_Balance'] > 0, 'Weighted_Average_Interest_Rate'].fillna(0)



    # Removing outliers:
    df = df[(df['Year'] >=2012) & (df['Year'] <=2022 )]
    table['Year earlier than 2012 and and later than 2022'] = len(df)
    df = df[(df['Weighted_Average_Interest_Rate'] >=0) & (df['Weighted_Average_Interest_Rate'] <=0.1 )]
    table['Weighted_Average_Interest_Rate less than 0 and larger than 0.1'] = len(df)
    df = df[df['Repayment_Loan_Balance'] >= 0]
    table['Repayment_Loan_Sum larger less than 0'] = len(df)
    df = df[df['Boligkreditt_Balance'] >= 0]
    table['Boligkreditt_Balance less than 0'] = len(df)
    df = df[(df['LTV'] >=0) & (df['LTV'] <=1 )]
    table['LTV less than 0 and larger than 1'] = len(df)
    df = df[(df['Age'] >= 18) & (df['Age'] <=78) ]
    table['Age younger than 18 and older than 78'] = len(df)
    df = df[df['Customer_Relationship_Duration'] >= 0]
    table['Customer_Relationship_Duration less than 0'] = len(df)
    df = df[df['Seniorlaan_Balance']>=0]
    table['Seniorlaan_Balance less than 0'] = len(df)
    df = df[(df['Other_External_Debt'] >= 0) & (df['Other_External_Debt'] <=50000000) ]
    table['Other_External_Debt less than 0 and larger than 50 000 000'] = len(df)
    df = df[df['Other_Internal_Debt']>=0]
    table['Other_Internal_Debt less than 0'] = len(df)
   

    # Create dummies for "Age" and "County"
    age = [0,39,64,150]
    selection = ['Age_Group_1', 'Age_Group_2', 'Age_Group_3']
    df['Age_Selection'] = pd.cut(df['Age'], bins = age, labels = selection)
    dummies = pd.get_dummies(df['Age_Selection'])
    df = pd.concat([df,dummies], axis=1)

    dummies_2 = pd.get_dummies(df['County'])
    df = pd.concat([df,dummies_2], axis=1)



    # Logtransformation of variables
    df['Age_Log'] = np.log(df['Age'])
    df['Customer_Relationship_Duration_Log'] = np.log(df['Customer_Relationship_Duration'])
    df['Customer_Relationship_Duration_Log'] = df['Customer_Relationship_Duration_Log'].replace([np.inf, -np.inf], 0)
    df['Repayment_Loan_Balance_Log'] = np.log(df['Repayment_Loan_Balance'])
    df['Repayment_Loan_Balance_Log'] = df['Repayment_Loan_Balance_Log'].replace([np.inf, -np.inf], 0)
    df['Boligkreditt_Balance_Log'] = np.log(df['Boligkreditt_Balance'])
    df['Boligkreditt_Balance_Log'] = df['Boligkreditt_Balance_Log'].replace([np.inf, -np.inf], 0)
    df['Boligkreditt_Credit_Limit_Log'] = np.log(df['Boligkreditt_Credit_Limit'])
    df['Boligkreditt_Credit_Limit_Log'] = df['Boligkreditt_Credit_Limit_Log'].replace([np.inf, -np.inf], 0)
    df['Seniorlaan_Balance_Log'] = np.log(df['Seniorlaan_Balance'])
    df['Seniorlaan_Balance_Log'] = df['Seniorlaan_Balance_Log'].replace([np.inf, -np.inf], 0)
    df['Other_Internal_Debt_Log'] = np.log(df['Other_Internal_Debt'])
    df['Other_Internal_Debt_Log'] = df['Other_Internal_Debt_Log'].replace([np.inf, -np.inf], 0)
    df['Other_External_Debt_Log'] = np.log(df['Other_External_Debt'])
    df['Other_External_Debt_Log'] = df['Other_External_Debt_Log'].replace([np.inf, -np.inf], 0)


    # Creating new variables:
    df['DTI'] = (df['Repayment_Loan_Balance'] + df['Boligkreditt_Balance'] + df['Seniorlaan_Balance'] + df['Other_Internal_Debt'] +df['Other_External_Debt'])/df['Income']
    df['DTI'] = np.where(np.isinf(df['DTI']), np.nan, df['DTI'])

    df['PD_LTV'] = df['PD']*df['LTV']
    df['PD_LTV'] = np.where(np.isinf(df['PD_LTV']), np.nan, df['PD_LTV'])


    # Remove observations for customers that have less than 6 observations
    customer_counts = df.groupby('Customer_ID').size()
    valid_customers = customer_counts[customer_counts >=6].index
    df = df[df['Customer_ID'].isin(valid_customers)]

    # Remove observations after churn for each churned customer
    df['Date'] = pd.to_datetime(df['Date_y'])
    df = df.sort_values(['Customer_ID','Date'])

    grouped = df.groupby('Customer_ID')

    def remove_rows(group):
        # Find the index of the first row with Churned_6_Lag = 1
        last_row = group['Churned_6_Lag']
        if (last_row == 1).any():
            last_row_index = last_row.idxmax()
            return group.loc[:last_row_index]
        else:
            return group

    df = grouped.apply(remove_rows)
    df = df.reset_index(drop=True)
    table['Observations after churn'] = len(df)
    tabledf = pd.DataFrame({'Data cleaning': list(table.keys()), 'Observations left': list(table.values())})
    tabledf = tabledf.set_index('Data cleaning')


    with open('data_cleaning.tex','w') as tf:
        tf.write(tabledf.to_latex())

    # Drop columns 
    df = df.drop(['Age_Selection','Age','County','KundeStart','KundeSlutt','Kjonn',
                  'Fodselsdato','Kredittkort','Innskudd','Kundefrafall','Municipality',
                  'Kommunenummer','Date_x','Date_y','Year','Month', 'MaanedTall', 'Churned_3_Lag',
                  'Churned','Repayment_Loan_Balance','Boligkreditt_Balance',
                  'Boligkreditt_Credit_Limit', 'Seniorlaan_Balance',
                  'Other_Internal_Debt', 'Other_External_Debt'],axis=1)
    

    # New column to indicate missing values
    df['Missing'] = np.where(df.isna().any(axis=1), np.nan, 1)

    # Empty column. Will be used in data_model
    df['Income_Log'] =''


    return df